# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE-MITMPROXY.
# Copyright (C) 2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Load/dump interaction files."""

from logging import getLogger
from os.path import expandvars
from pathlib import Path
from pprint import pformat
from re import compile
from threading import Timer
from typing import Any, Dict, List, Optional, Pattern, Union

import requests
from yaml import dump as yaml_dump
from yaml import load as yaml_load

from .http import MITMRequest, MITMResponse, response_to_string


logger = getLogger(__name__)


class Interaction:
    DEFAULT_EXACT_MATCH_FIELDS: List[str] = ['url', 'method', 'body']
    DEFAULT_REGEX_MATCH_FIELDS: Dict[str, Pattern[str]] = {}
    DEFAULT_CALLBACK_DELAY = 0.5

    DEFAULT_NAME_PATTERN = 'interaction_{}'
    DEFAULT_NAME_MATCH_REGEX = compile(r'^interaction_(\d+)$')

    def __init__(
        self,
        name: str,
        request: MITMRequest,
        response: MITMResponse,
        match: Optional[dict] = None,
        callbacks: Optional[List[dict]] = None,
    ) -> None:
        self.name = name
        self.request = request
        self.response = response
        self.match = match or {}
        self.callbacks = callbacks or []

    @classmethod
    def from_file(cls, interaction_file: Optional[Path]) -> 'Interaction':
        interaction_string = interaction_file.read_text()  # type: ignore
        interaction_dict = yaml_load(interaction_string)

        return cls(
            name=interaction_file.stem,  # type: ignore
            request=MITMRequest.from_dict(interaction_dict['request']),
            response=MITMResponse.from_dict(interaction_dict['response']),
            match=interaction_dict.get('match'),
            callbacks=interaction_dict.get('callbacks'),
        )

    def to_dict(self) -> dict:
        serialized_interaction: Dict[str, Any] = {
            'request': self.request.to_dict(),
            'response': self.response.to_dict(),
            'match': self.match,
            'callbacks': self.callbacks,
        }

        return serialized_interaction

    @property
    def exact_match_fields(self) -> List[str]:
        if not self.match:
            return self.DEFAULT_EXACT_MATCH_FIELDS

        try:
            return self.match['exact']
        except KeyError:
            return []

    @property
    def regex_match_fields(self) -> Dict[str, Pattern[str]]:
        if not self.match:
            return self.DEFAULT_REGEX_MATCH_FIELDS

        try:
            regexes = self.match['regex']
            return {field: compile(regex) for field, regex in regexes.items()}
        except KeyError:
            return {}

    def _matches_by_exact_rules(self, request: MITMRequest) -> bool:
        for match_on in self.exact_match_fields:
            if self.request[match_on] != request[match_on]:
                return False

        return True

    def _matches_by_regex_rules(self, request: MITMRequest) -> bool:
        for match_on, regex in self.regex_match_fields.items():
            if not regex.match(request[match_on]):
                return False

        return True

    def matches_request(self, request: MITMRequest) -> bool:
        return self._matches_by_exact_rules(request) and self._matches_by_regex_rules(request)

    @staticmethod
    def execute_callback(request: MITMRequest, delay: Union[int, float]):
        def execute_request(_request: MITMRequest):
            request_params = dict(
                method=request.method,
                url=expandvars(request.url),
                data=request.body,
                headers={
                    key: expandvars(request.headers[key])
                    for key in request.headers.keys()
                },
                timeout=10,
            )
            logger.warning("Executing callback:\n%s", pformat(request_params))

            response = requests.request(**request_params)
            if not response.ok:
                logger.error("Error executing callback:\n%s", response_to_string(response))

        timer = Timer(delay, execute_request, args=[request])
        timer.start()

    def execute_callbacks(self):
        for callback in self.callbacks:
            request = MITMRequest.from_dict(callback['request'])
            self.execute_callback(
                request=request,
                delay=callback.get('delay', self.DEFAULT_CALLBACK_DELAY)
            )

    @classmethod
    def get_next_sequence_number_in_dir(cls, directory: Path) -> int:
        def _next_sequence_number_after_file(path):
            if not path.is_file() or path.suffix != '.yaml':
                return 0

            seq_number_match = cls.DEFAULT_NAME_MATCH_REGEX.match(path.stem)
            if not seq_number_match:
                return 0

            cur_seq_number = int(seq_number_match.group(1))
            return cur_seq_number + 1

        next_seq_number = 0
        for interaction_path in directory.iterdir():
            candidate_next_seq_number = _next_sequence_number_after_file(interaction_path)
            next_seq_number = max(next_seq_number, candidate_next_seq_number)

        return next_seq_number

    @classmethod
    def next_in_dir(
        cls,
        directory: Path,
        request: MITMRequest,
        response: MITMResponse
    ) -> 'Interaction':
        """Create a new interaction with a name taking from next available in directory."""
        sequence_number = cls.get_next_sequence_number_in_dir(directory)
        new_name = cls.DEFAULT_NAME_PATTERN.format(sequence_number)
        return Interaction(name=new_name, request=request, response=response)

    def save_in_dir(self, directory: Path):
        """Save the interaction to a file."""
        output_path = directory / f'{self.name}.yaml'
        output_path.write_text(yaml_dump(self.to_dict()))

    def __repr__(self):
        return f'Interaction(name={self.name!r}, request={self.request!r}, ' \
            f'response={self.response!r}, match={self.match!r}, callbacks={self.callbacks!r})'

    def __eq__(self, other):
        return (
            self.name == other.name
            and self.request == other.request
            and self.response == other.response
            and self.match == other.match
            and self.callbacks == other.callbacks
        )



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE-MITMPROXY.
# Copyright (C) 2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Dispatcher forwards requests to Services."""

from logging import getLogger
from typing import List, Type, cast

from mitmproxy.http import HTTPFlow, HTTPResponse

from .errors import DoNotIntercept, NoServicesForRequest
from .http import MITMRequest, MITMResponse
from .services import (
    ArxivService,
    BaseService,
    LegacyService,
    ManagementService,
    RTService,
    WhitelistService
)


logger = getLogger(__name__)


class Dispatcher:
    SERVICE_LIST: List[Type[BaseService]] = [
        ArxivService,
        LegacyService,
        RTService,
        WhitelistService,
    ]

    def __init__(self) -> None:
        self.services = [service_class() for service_class in self.SERVICE_LIST]
        mgmt_service = ManagementService(self.services)
        self.services = [cast(BaseService, mgmt_service)] + self.services

    def find_service_for_request(self, request: MITMRequest) -> BaseService:
        for service in self.services:
            if service.handles_request(request):
                return service
        raise NoServicesForRequest(request)

    def process_request(self, request: MITMRequest) -> MITMResponse:
        """Perform operations and give response."""
        service = self.find_service_for_request(request)
        return service.process_request(request)

    def process_response(self, request: MITMRequest, response: MITMResponse):
        """Hook for live responses."""
        service = self.find_service_for_request(request)
        service.process_response(request=request, response=response)

    def request(self, flow: HTTPFlow):
        """MITMProxy addon event interface for outgoing request."""
        try:
            request = MITMRequest.from_mitmproxy(flow.request)
            response = self.process_request(request).to_mitmproxy()
            flow.response = response
        except DoNotIntercept as e:
            # Let the request pass through, by not interrupting the flow, but log it
            logger.warning(str(e))
        except Exception as e:
            flow.response = HTTPResponse.make(
                status_code=getattr(e, 'http_status_code', 500),
                content=str(e),
                headers={'Content-Type': 'text/plain'}
            )

    def response(self, flow: HTTPFlow):
        if self.is_flow_passed_through(flow):
            request = MITMRequest.from_mitmproxy(flow.request)
            response = MITMResponse.from_mitmproxy(flow.response)
            self.process_response(request, response)

    @staticmethod
    def is_flow_passed_through(flow: HTTPFlow) -> bool:
        return flow.server_conn.connected()



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE-MITMPROXY.
# Copyright (C) 2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Service used to orchestrate fake services."""

from json import JSONDecodeError
from json import dumps as json_dumps
from json import loads as json_loads
from os import environ
from pathlib import Path
from re import compile
from typing import Dict, List, Match, Optional, Union, cast
from urllib.parse import urlparse

from autosemver.packaging import get_current_version

from ..errors import InvalidRequest, RequestNotHandledInService, ServiceNotFound
from ..http import MITMHeaders, MITMRequest, MITMResponse
from ..services import BaseService


class ManagementService(BaseService):
    SERVICE_HOSTS = ['mitm-manager.local']
    INTERACTIONS_ENDPOINT = compile(r'/service/(\w+)/interactions')

    def __init__(self, services: List[BaseService]) -> None:
        super(ManagementService, self).__init__()

        self.services = services
        self.config = {
            'active_scenario': 'default',
        }
        self.is_recording = False
        self.propagate_option_changes()

    def get_active_scenario(self):
        return self.config.get('active_scenario', 'default')

    def process_request(self, request: MITMRequest) -> MITMResponse:
        parsed_url = urlparse(request.url)
        path = parsed_url.path
        method = request.method

        if path == '/services' and method == 'GET':
            return self.build_response(200, self.get_services())
        elif self.INTERACTIONS_ENDPOINT.match(path) and method == 'GET':
            match = cast(Match[str], self.INTERACTIONS_ENDPOINT.match(path))
            service_name = match.group(1)
            return self.build_response(200, self.get_service_interactions(service_name))
        elif path == '/scenarios' and method == 'GET':
            return self.build_response(200, self.get_scenarios())
        elif path == '/config' and method == 'GET':
            return self.build_response(200, self.get_config())
        elif path == '/config' and method == 'PUT':
            return self.build_response(204, self.put_config(request))
        elif path == '/config' and method == 'POST':
            return self.build_response(201, self.post_config(request))
        elif path == '/record' and method == 'PUT':
            return self.build_response(204, self.set_recording(request))
        elif path == '/record' and method == 'POST':
            return self.build_response(201, self.set_recording(request))

        raise RequestNotHandledInService(self.name, request)

    def get_services(self) -> dict:
        return {
            idx: {
                'class': service.name,
                'service_hosts': service.SERVICE_HOSTS,
            }
            for idx, service in enumerate([cast(BaseService, self)] + self.services)
        }

    def get_scenarios(self) -> dict:
        path = Path(environ.get('SCENARIOS_PATH', './scenarios/'))
        response: Dict[str, Dict[str, Dict[str, List[str]]]] = {
            scenario.name: {
                'responses': {
                    service.name: [
                        response_file.name for response_file in sorted(service.iterdir())
                        if response_file.is_file() and response_file.suffix == '.yaml'
                    ]
                    for service in scenario.iterdir() if service.is_dir()
                }
            }
            for scenario in path.iterdir() if scenario.is_dir()
        }

        return response

    def get_config(self) -> dict:
        return self.config

    def put_config(self, request: MITMRequest):
        try:
            config_update = json_loads(request.body)
            self.config.update(config_update)
            self.propagate_option_changes()
        except (JSONDecodeError, ValueError):
            raise InvalidRequest(self.name, request)

    def post_config(self, request: MITMRequest):
        try:
            self.config = json_loads(request.body)
            self.propagate_option_changes()
        except JSONDecodeError:
            raise InvalidRequest(self.name, request)

    def set_recording(self, request: MITMRequest):
        try:
            recording_opts = json_loads(request.body)
            self.is_recording = recording_opts['enable']
            self.propagate_option_changes()
            return {
                'enabled': self.is_recording,
            }
        except (JSONDecodeError, KeyError, TypeError):
            raise InvalidRequest(self.name, request)

    def build_response(self, code: int, json_message: Optional[Union[dict, list]]) -> MITMResponse:
        try:
            body = json_dumps(json_message, indent=2)
        except JSONDecodeError:
            body = ''

        return MITMResponse(
            status_code=code,
            body=body,
            headers=MITMHeaders({
                'Content-Type': ['application/json; encoding=UTF-8'],
                'Server': [
                    'inspire-mitmproxy/' + get_current_version(project_name='inspire_mitmproxy')
                ]
            }),
        )

    def get_service_interactions(self, service_name) -> dict:
        for service in self.services:
            if service.name == service_name:
                return service.interactions_replayed.get(self.get_active_scenario(), {})

        raise ServiceNotFound(service_name)

    def propagate_option_changes(self):
        """On change of config, propagate relevant information to services."""
        for service in self.services:
            service.active_scenario = self.get_active_scenario()
            service.is_recording = self.is_recording



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE-MITMPROXY.
# Copyright (C) 2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Wrappers for HTTP objects."""

from cgi import parse_header
from copy import copy, deepcopy
from socket import getservbyname
from typing import Any, Dict, KeysView, List, Optional, Union
from urllib.parse import urlparse

import requests
from mitmproxy.http import HTTPRequest, HTTPResponse
from mitmproxy.net.http.headers import Headers
from mitmproxy.net.http.status_codes import RESPONSES


def encoding_by_header(headers: 'MITMHeaders') -> str:
    """Extract charset param from Content-Type or Accept headers"""
    try:
        content_type = headers['Content-Type']
        _, params = parse_header(content_type)
        return params['charset']
    except KeyError:
        return 'utf-8'


def response_to_string(res: requests.Response) -> str:
    """
    :param res: :class:`requests.Response` object
    Parse the given request and generate an informative string from it
    """
    if 'Authorization' in res.request.headers:
        res.request.headers['Authorization'] = "*****"
    return """
####################################
url = %s
headers = %s
-------- data sent -----------------
%s
------------------------------------
@@@@@ response @@@@@@@@@@@@@@@@
headers = %s
code = %d
reason = %s
--------- data received ------------
%s
------------------------------------
####################################
""" % (res.url,
        str(res.request.headers),
        res.request.body,
        res.headers,
        res.status_code,
        res.reason,
        res.text)


class MITMHeaders:
    def __init__(self, headers: Dict[str, List[str]]) -> None:
        self.headers: Dict[str, List[str]] = {}
        for header_name, header_value in headers.items():
            header_name = header_name.title()
            self.headers[header_name] = copy(header_value)

    @classmethod
    def from_dict(cls, headers_dict: Dict[str, List[str]]) -> 'MITMHeaders':
        return cls(headers=headers_dict)

    @classmethod
    def from_mitmproxy(cls, headers: Headers) -> 'MITMHeaders':
        header_dict: Dict[str, List[str]] = {}

        for key, value in headers.fields:
            key, value = key.decode('ascii'), value.decode('ascii')

            if key in header_dict:
                header_dict[key].append(value)
            else:
                header_dict[key] = [value]

        return cls(headers=header_dict)

    def to_dict(self) -> Dict[str, List[str]]:
        return deepcopy(self.headers)

    def to_mitmproxy(self) -> Headers:
        fields = []

        for key, values in self.headers.items():
            for value in values:
                fields.append(
                    (key.encode('ascii'), value.encode('ascii'))
                )

        return Headers(fields=fields)

    def keys(self) -> KeysView[str]:
        return self.headers.keys()

    def __getitem__(self, item: str) -> str:
        try:
            return self.headers[item.title()][0]
        except IndexError:
            raise KeyError(item.title())

    def __eq__(self, other) -> bool:
        return self.headers == other.headers

    def __repr__(self):
        return f'MITMHeaders(headers={self.headers!r})'


class MITMRequest:
    def __init__(
        self,
        url: str,
        method: str = 'GET',
        body: Optional[Union[str, bytes]] = None,
        headers: Optional[MITMHeaders] = None,
        original_encoding: Optional[str] = None,
        http_version: Optional[str] = None,
    ) -> None:
        self.url = url
        self.method = method
        self.headers = headers or MITMHeaders({})
        self.http_version = http_version or 'HTTP/1.1'
        self.original_encoding = original_encoding or encoding_by_header(self.headers)

        if isinstance(body, str):
            self.body = body.encode(self.original_encoding)
        elif isinstance(body, bytes):
            self.body = body
        else:
            self.body = b''

    @classmethod
    def from_mitmproxy(cls, request: HTTPRequest) -> 'MITMRequest':
        encoding = encoding_by_header(MITMHeaders.from_mitmproxy(request.headers))

        return cls(
            url=request.url,
            method=request.method,
            body=request.content,
            headers=MITMHeaders.from_mitmproxy(request.headers),
            original_encoding=encoding,
            http_version=request.http_version,
        )

    @classmethod
    def from_dict(cls, request: Dict[str, Any]) -> 'MITMRequest':
        encoding = encoding_by_header(MITMHeaders.from_dict(request['headers']))

        return cls(
            url=request['url'],
            method=request['method'],
            body=request['body'],
            headers=MITMHeaders.from_dict(request['headers']),
            original_encoding=encoding,
        )

    def to_mitmproxy(self) -> HTTPRequest:
        parsed_url = urlparse(self.url)

        return HTTPRequest(
            first_line_format='absolute',
            method=self.method,
            scheme=parsed_url.scheme,
            host=parsed_url.hostname,
            port=parsed_url.port or getservbyname(parsed_url.scheme),
            path=parsed_url.path,
            http_version=self.http_version,
            headers=self.headers.to_mitmproxy(),
            content=self.body,
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            'method': self.method,
            'url': self.url,
            'body': self.body.decode(self.original_encoding),
            'headers': self.headers.to_dict(),
        }

    def __eq__(self, other) -> bool:
        return (
            self.url == other.url
            and self.method == other.method
            and self.body == other.body
            and self.headers == other.headers
        )

    def __repr__(self):
        return f'MITMRequest({self.url!r}, {self.method!r}, ' \
            f'headers={self.headers!r}, body={self.body!r})'

    def __getitem__(self, field: str):
        return getattr(self, field)


class MITMResponse:
    def __init__(
        self,
        status_code: int = 200,
        status_message: Optional[str] = None,
        body: Optional[Union[str, bytes]] = None,
        headers: Optional[MITMHeaders] = None,
        original_encoding: Optional[str] = None,
        http_version: Optional[str] = None,
    ) -> None:
        self.status_code = status_code
        self.status_message = status_message or RESPONSES[status_code]
        self.headers = headers or MITMHeaders({})
        self.http_version = http_version or 'HTTP/1.1'
        self.original_encoding = original_encoding or encoding_by_header(self.headers)

        if isinstance(body, str):
            self.body = body.encode(self.original_encoding)
        elif isinstance(body, bytes):
            self.body = body
        else:
            self.body = b''

    @classmethod
    def from_mitmproxy(cls, response: HTTPResponse) -> 'MITMResponse':
        encoding = encoding_by_header(MITMHeaders.from_mitmproxy(response.headers))

        return cls(
            status_code=response.status_code,
            status_message=response.reason,
            body=response.content.decode(encoding),
            headers=MITMHeaders.from_mitmproxy(response.headers),
            original_encoding=encoding,
            http_version=response.http_version,
        )

    @classmethod
    def from_dict(cls, response: Dict[str, Any]) -> 'MITMResponse':
        return cls(
            status_code=response['status']['code'],
            status_message=response['status']['message'],
            body=response['body'],
            headers=MITMHeaders.from_dict(response['headers']),
        )

    def to_mitmproxy(self) -> HTTPResponse:
        return HTTPResponse(
            http_version='HTTP/1.1',
            status_code=self.status_code,
            reason=self.status_message,
            headers=self.headers.to_mitmproxy(),
            content=self.body,
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            'status': {
                'code': self.status_code,
                'message': self.status_message,
            },
            'body': self.body.decode(self.original_encoding),
            'headers': self.headers.to_dict(),
        }

    def __eq__(self, other) -> bool:
        return (
            self.status_code == other.status_code
            and self.body == other.body
            and self.headers == other.headers
        )

    def __repr__(self):
        return f'MITMResponse({self.status_code!r}, {self.status_message!r}, ' \
            f'headers={self.headers!r}, body={self.body!r})'



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE-MITMROXY.
# Copyright (C) 2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Service which allows all requests outside"""
import os

from ..errors import DoNotIntercept
from ..http import MITMRequest, MITMResponse
from ..services import BaseService


class WhitelistService(BaseService):
    DEFAULT_HOSTS: str = 'test-indexer test-scrapyd test-web-e2e.local'

    def __init__(self, *args, **kwargs):
        self.SERVICE_HOSTS = os.environ.get(
            'MITM_PROXY_WHITELIST',
            self.DEFAULT_HOSTS
        ).split()
        super().__init__(*args, **kwargs)

    def process_request(self, request: MITMRequest):
        raise DoNotIntercept(self.name, request)

    def process_response(self, request: MITMRequest, response: MITMResponse):
        pass



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE-MITMPROXY.
# Copyright (C) 2018 CERN.
#
# INSPIRE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE. If not, see <http://www.gnu.org/licenses/>.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Base for fake services."""

from os import environ
from pathlib import Path
from typing import Any, Dict, List
from urllib.parse import splitport  # type: ignore
from urllib.parse import urlparse

from ..errors import DoNotIntercept, NoMatchingRecording, ScenarioNotInService
from ..http import MITMRequest, MITMResponse
from ..interaction import Interaction


class BaseService:
    """Mocked service base."""
    SERVICE_HOSTS: List[str] = []

    def __init__(self):
        self.active_scenario: str = 'default'
        self.interactions_replayed: Dict[str, Dict[str, Dict[str, Any]]] = {}
        self.is_recording = False

    @property
    def name(self):
        return type(self).__name__

    def handles_request(self, request: MITMRequest) -> bool:
        """Can this service handle the request?"""
        try:
            host = splitport(request.headers['Host'])[0]
        except (TypeError, KeyError):
            host = urlparse(request.url).hostname

        return host in self.SERVICE_HOSTS

    def _get_first_matching_interaction(self, request):
        for interaction in self.get_interactions_for_active_scenario():
            if interaction.matches_request(request):
                return interaction

    def _raise_do_not_intercept_if_recording(self, request):
        if self.is_recording:
            raise DoNotIntercept(self.name, request)

    def process_request(self, request):
        try:
            matched_interaction = self._get_first_matching_interaction(request)
        except ScenarioNotInService:
            self._raise_do_not_intercept_if_recording(request)
            raise

        if matched_interaction is None:
            self._raise_do_not_intercept_if_recording(request)
            raise NoMatchingRecording(self.name, request)

        response = matched_interaction.response
        matched_interaction.execute_callbacks()
        self.increment_interaction_count(matched_interaction.name)
        return response

    def process_response(self, request: MITMRequest, response: MITMResponse):
        """Perform operations based on live response."""
        if not self.is_recording:
            return

        current_scenario_dir = self.get_path_for_active_scenario_dir(create=True)
        interaction = Interaction.next_in_dir(
            directory=current_scenario_dir,
            request=request,
            response=response,
        )
        interaction.save_in_dir(current_scenario_dir)

    def increment_interaction_count(self, interaction_name: str):
        try:
            self.interactions_replayed[self.active_scenario][interaction_name]['num_calls'] += 1
        except KeyError:
            self.interactions_replayed.setdefault(
                self.active_scenario,
                {},
            ).setdefault(
                interaction_name,
                {'num_calls': 1},
            )

    def get_path_for_active_scenario_dir(self, create=False) -> Path:
        scenarios_path = Path(environ.get('SCENARIOS_PATH', './scenarios/'))
        interactions_dir = scenarios_path / self.active_scenario / self.name

        if create:
            interactions_dir.mkdir(parents=True, exist_ok=True)

        return interactions_dir

    def get_interactions_in_scenario(self, scenario_path: Path) -> List[Interaction]:
        return [
            Interaction.from_file(interaction_file=interaction_path)
            for interaction_path in sorted(scenario_path.iterdir())
            if interaction_path.is_file() and interaction_path.suffix == '.yaml'
        ]

    def get_interactions_for_active_scenario(self) -> List[Interaction]:
        """Get a list of scenarios"""
        scenario_dir = self.get_path_for_active_scenario_dir(create=False)

        if not scenario_dir.exists():
            raise ScenarioNotInService(self.name, self.active_scenario)

        return self.get_interactions_in_scenario(scenario_dir)
